/**
 * Temporal Quantified fuzzy DL's
 */
package reasoner;

import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;
import util.LinearInterpolation;
import util.Point;

/** 
 * This class represents a logic Concept as a FuzzySet of Individual
 * @author marco furlan
 */
public class Concept implements FuzzySet {
	/**
	 * the map of EventInd and values
	 */	
	protected Map<Individual, Double> events;
	/**
	 * the map of ULineInd and values
	 */
	protected Map<Individual, Double> uLineMap;	
	/**
	 * the map of StaticInd and values
	 */
	protected Map<Individual, Double> staticMap;	
	/**
	 * the sometime map
	 */	
	private Map<Individual, Double> sometimeMap;
	/**
	 * the always map
	 */	
	private Map<Individual, Double> alwaysMap;

	/**
	 * the concept name
	 */
	private String name;
	/**
	 * the Knowledge Base
	 */
	protected static KnowledgeBase kb;
	/**
	 * the constant _now_
	 */
	protected static double NOW;
	/**
	 * the constant _self_
	 */
	protected static double SELF;
	/**
	 * utility formatter
	 */
	protected NumberFormat formatter; 

	/**
	 * Constructs an empty Concept, with a given name and in a given KnowledgeBase
	 * @param name the name
	 * @param kb the KnowledgeBase
	 */
	public Concept(String name, KnowledgeBase kb) {
		setKb(kb);
		setName(name);
		setEvents(new TreeMap<Individual, Double>());			
		setULineMap(new TreeMap<Individual, Double>());
		setStaticMap(new TreeMap<Individual, Double>());
		setSometimeMap(new TreeMap<Individual, Double>());
		setAlwaysMap(new TreeMap<Individual, Double>());
		getKb().addConcept(this);
		setFormatter(3);
	}

	/**
	 * Constructs a Concept from the given map of EventInds and values and with
	 * a given name and in a given knowledge base
	 * @param ev the given map of EventInds and values
	 * @param name the name
	 * @param kb the knowledge base
	 */
	public Concept(Map<Individual, Double> ev, String name, KnowledgeBase kb) {
		setKb(kb);
		setEvents(ev);
		setULineMap(new TreeMap<Individual, Double>());
		setStaticMap(new TreeMap<Individual, Double>());
		setName(name);
		getKb().addConcept(this);
		setFormatter(3);		
	}
	private void setFormatter(int number){
		formatter = NumberFormat.getNumberInstance();
		formatter.setMaximumFractionDigits(number);
	}
	
	/**
	 * Gives this Concept name
	 * @return the name
	 */
	public String getName() {
		return name;
	}

	/**
	 * Sets this Concept name
	 * @param name the name to set
	 */
	public void setName(String name) {
		this.name = name;
	}

	/**
	 * Returns the membership value of the given Individual on the
	 * logical expression [sometimeA(ind)], where A is this Concept, using
	 * the 'sup' semantics.
	 * @param ind the given Individual
	 * @return the membership value
	 */
	public double sometime (Individual ind) {
		ULineEvent ul;
		Individual old, ind2;
		double val, t1, t2;
		ul = ((DynamicInd)ind).getULineEvent();
		t1 = ind.getTime();
		old = this.getLastSometime(ul);
		if (ind instanceof ULineInd)
			t1=ul.getLast().getTime();		
		if(old == null){
			ind2 = ul.getFirst();								
			val = ind2.degree(this);
		}else{
			ind2 = old;
			val=this.getSometimeMap().get(ind2);
		}	
		t2 = ind2.getTime();
		if (t1 < t2){			//from the beginning
			ind2 = ul.getFirst();						
			val = ind2.degree(this);
			t2 = ind2.getTime();
		}		
		while (t2<t1){
			ind2 = ul.getNext(ind2);
			if (ind2!=null){
				val = Math.max(val, ind2.degree(this));
				t2 = ind2.getTime();
			}
		}
		if (t2==t1){
			if (old != null && !old.equals(ind2)){
				this.getSometimeMap().remove(old);
				this.getSometimeMap().put(ind2, val);
			}else if (old == null)
				this.getSometimeMap().put(ind2, val);
			return val;
		}		
		return 0.0;
	}

	/**
	 * Returns the membership value of the given Individual on the
	 * logical expression [alwaysA(ind)], where A is this Concept, using
	 * the 'inf' semantics.
	 * @param ind the given Individual
	 * @return the membership value
	 */
	public double always (Individual ind) {
		ULineEvent ul;
		Individual old, ind2;
		double val, t1, t2;
		ul = ((DynamicInd)ind).getULineEvent();
		t1 = ind.getTime();
		old = this.getLastAlways(ul);
//		if (t1 == getSELF())
		if (ind instanceof ULineInd)
			t1=ul.getLast().getTime();	
		if(old == null){
			ind2 = ul.getFirst();								
			val = ind2.degree(this);
		}else{
			ind2 = old;
			val=this.getAlwaysMap().get(ind2);
		}	
		t2 = ind2.getTime();
		if (t1 < t2){			//from the beginning
			ind2 = ul.getFirst();						
			val = ind2.degree(this);
			t2 = ind2.getTime();
		}		
		while (t2<t1){
			ind2 = ul.getNext(ind2);
			if (ind2!=null){
				val = Math.min(val, ind2.degree(this));
				t2 = ind2.getTime();
			}
		}
		if (t2==t1){
			if (old != null && !old.equals(ind2)){
				this.getAlwaysMap().remove(old);
				this.getAlwaysMap().put(ind2, val);
			}else if (old == null)
				this.getAlwaysMap().put(ind2, val);
			return val;
		}		
		return 0.0;
	}

	/**
	 * Returns the last Individual, belonging to the given ULineEvent, 
	 * for which the sometime value has been evaluated.
	 * @param ul the given ULineEvent
	 * @return the last Individual
	 */
	private Individual getLastSometime(ULineEvent ul){
		Set<Individual> keys = this.getSometimeMap().keySet();
		Iterator<Individual> it = keys.iterator();
		while (it.hasNext()){
			Individual ind = (Individual)it.next();
			if (((DynamicInd)ind).getULineEvent().equals(ul))
				return ind;
		}		
		return null;
	}

	/**
	 * Returns the last Individual, belonging to the given ULineEvent,
	 * for which the always value has been evaluated.
	 * @param ul the given ULineEvent
	 * @return the last Individual
	 */
	private Individual getLastAlways(ULineEvent ul){
		Set<Individual> keys = this.getAlwaysMap().keySet();
		Iterator<Individual> it = keys.iterator();
		while (it.hasNext()){
			Individual ind = (Individual)it.next();
			if (((DynamicInd)ind).getULineEvent().equals(ul))
				return ind;
		}		
		return null;
	}

	/**
	 * Executes the evaluation of the logical expression [A until B(ind)],
	 * where A is this Concept, B is a given other Concept, ind is the Individual for which
	 * expression has to be evaluated; lastUntil is the given expression value for
	 * the last (previous) Individual with respect to ind.
	 * Operator Until has the weak and inclusive definition.
	 * Algorithm is incremental and dynamic: has last values memory. 
	 * @param b the Concept B
	 * @param ind the Individual
	 * @param lastUntil the last value
	 * @return the until evaluation result
	 */
	public double until (Concept b, Individual ind, double lastUntil){
		System.out.println(this.getName()+"Until"+b.getName()+"("+ind.getName()+"), with lastUntil: "+formatter.format(lastUntil));
		double some =this.sometime(ind);
		double bval = b.get(ind);
		System.out.println("sometime"+this.getName()+"("+ind.getName()+"): "+ formatter.format(some));
		System.out.println(b.getName()+"("+ ind.getName()+"): "+formatter.format(bval));
		double newUntil =Math.max(lastUntil, Math.min(bval,some ));
		System.out.println("new Until: " + formatter.format(newUntil));
		return newUntil;
	}

	/**
	 * Executes the evaluation of the logical expression [A until B(ind)],
	 * where A is this Concept, B is a given Concept, ind is the Individual for which
	 * expression has to be evaluted.
	 * Operator Until has the weak and inclusive definition. 
	 * Algorithm is incremental and not dynamic: without memory, scan all the 
	 * ULineEvent history. 
	 * @param b the Concept B
	 * @param ind the Individual	
	 * @return the until evaluation result
	 */
	public double until (Concept b, Individual ind){
		System.out.println(this.getName() +"Until" +b.getName()+"("+ind.getName()+"):...");
		double until = 0.0;
		System.out.println("until: " + formatter.format(until));
		double time = ind.getTime();
		System.out.println("time: " + time);
		ULineEvent ul = ((DynamicInd)ind).getULineEvent();
		Individual xo = ul.getFirst();
		System.out.println("first_xo: " + xo.getName());
		double time0 = xo.getTime();
		System.out.println("time0: " + time0);
//		finch xo raggiunge ind
		while (time0 <= time || 						
//				oppure, se ind  il rappresentante di ul, xo raggiunge l'ultimo (calcolo su tutto ul)
				( ind instanceof ULineInd && time0 != ul.getLast().getTime())){
			if (!(xo instanceof ULineInd)) {
				until = this.until(b, xo, until);	
				System.out.println("until: " + formatter.format(until));
			}			
			xo = ul.getNext(xo);
			if (xo==null)
				break;
			System.out.println("xo: " + xo.getName());
			time0 = xo.getTime();
			System.out.println("time0: " + time0);
		}			
		return until;
	}
	/**
	 * Executes the evaluation of the logical expression [A since B(ind)],
	 * where A is this Concept, B is a given Concept, ind is the Individual for which
	 * expression has to be evaluted.
	 * Operator Since has the weak and inclusive definition. 
	 * Algorithm is incremental and not dynamic: without memory, scan all the 
	 * ULineEvent history. 
	 * @param b the Concept B
	 * @param ind the Individual	
	 * @return the since evaluation result
	 */
	public double since (Concept b, Individual ind){
		System.out.println(this.getName() +"Since" +b.getName()+"("+ind.getName()+"):...");
		double since = 0.0;
		System.out.println("since: " + formatter.format(since));
		double time = ind.getTime();
		System.out.println("time: " + time);
		ULineEvent ul = ((DynamicInd)ind).getULineEvent();
		Individual xo = ul.getFirst();
		System.out.println("first_xo: " + xo.getName());
		double time0 = xo.getTime();
		System.out.println("time0: " + time0);
//		finch xo raggiunge ind
		while (time0 <= time || 						
//				oppure, se ind  il rappresentante di ul, xo raggiunge l'ultimo (calcolo su tutto ul)
				( ind instanceof ULineInd && time0 != ul.getLast().getTime())){
			if (!(xo instanceof ULineInd)) {				
				this.getSometimeMap().put(xo, xo.degree(this));
				since = this.since(b, ind, since);	///////
				System.out.println("since: " + formatter.format(since));
			}			
			xo = ul.getNext(xo);
			if (xo==null)
				break;
			System.out.println("xo: " + xo.getName());
			time0 = xo.getTime();
			System.out.println("time0: " + time0);
		}			
		return since;
	}
	/**
	 * Executes the evaluation of the logical expression [A since B(ind)],
	 * where A is this Concept, B is a given other Concept, ind is the Individual for which
	 * expression has to be evaluated; lastSince is the given expression value for
	 * the last (previous) Individual with respect to ind.
	 * Operator Since has the weak and inclusive definition.
	 * Algorithm is incremental and dynamic: has last values memory. 
	 * @param b the Concept B
	 * @param ind the Individual
	 * @param lastSince the last value
	 * @return the since evaluation result
	 */
	public double since (Concept b, Individual ind, double lastSince){
		System.out.println(this.getName()+"Since"+b.getName()+"("+ind.getName()+"), with lastSince: "+formatter.format(lastSince));
		double some =this.sometime(ind);
		double bval = b.get(ind);
		System.out.println("sometime"+this.getName()+"("+ind.getName()+"): "+ formatter.format(some));
		System.out.println(b.getName()+"("+ ind.getName()+"): "+formatter.format(bval));
		double newSince =Math.max(lastSince, Math.min(bval,some ));
		System.out.println("new Since: " + formatter.format(newSince));
		return newSince;
	}
	/**
	 * Add an Individual to this Concept, with a given membership value
	 * @param i the Individual
	 * @param d the membership value
	 * @return this Concept after adding Individual
	 */
	public Concept add(Individual i, Double d){
		if (i instanceof ULineInd){
			getULineMap().put(i, d);
		}
		else{
			getEvents().put(i, d);	
		}
		return this;
	}


	
	/**	 
	 * Gives this Concept cardinality at an alpha level. If mode = "any", evaluation 
	 * on all Events; if mode = "now", evaluation on now Events;
	 * if mode = "ul", evaluation on ULines
	 * @param alpha the alpha level
	 * @param mode the mode
	 * @return the integer number cardinality of the alpha-set
	 */
	public Integer card(Double alpha, String mode) {
		Set<Individual> keys =null;
		Iterator<Individual> iter;
		Double time = null;
		int card=0;
		if (mode.equals("any"))time=null;
		else if (mode.equals("now"))time=NOW;
		else if (mode.equals("ul"))time=SELF;
		else {
			System.out.println("error: set mode = \"any\" or \"now\" or \"ul\"" );
			return 0;
		}
		if(time==null ||time.equals(NOW))
			keys =this.getEvKeys();			//to be expanded for Statics
		else if(time.equals(SELF))
			keys =this.getULineMap().keySet();
		iter= keys.iterator();
		while(iter.hasNext()){
			Individual key =(Individual) iter.next();
			if (time == null || (((Double)key.getTime()).equals(time))){
				Double value=this.get(key);
				if (value>0.0 && value>=alpha)
					card++;
			}		
		}	
		return card;
	}
	/**
	 * Constructs an ArrayList of Points, containing Concept values, and,
	 * on it, produces a linear interpolation.
	 * @return a  new Point as the couple [angular coefficient, intercept]
	 */
	public Point interpole() {
		Set<Individual> keys =this.getEvents().keySet();
		Iterator<Individual> iter= keys.iterator();
		Double angCoeff, q, time, totaltime;
		Point p;
		ArrayList<Point> al = new ArrayList<Point>();
		totaltime = 20.0;					//to be set
		double size = keys.size();
		double unit = totaltime / size;
		time = 0.0;
		while(iter.hasNext()){
			Individual key =(Individual) iter.next();		
			Double value=this.getEvents().get(key);
			p = new Point(time, value);	
			al.add(p);
			time += unit;
		}	
		LinearInterpolation li = new LinearInterpolation(al);
		li.interpole();
		angCoeff = li.getM();
		q = li.getQ();
		return new Point(angCoeff, q);		
	}
	
	/**
	 * @return a TreeSet<Individual> as the this Concept EventInd set.  
	 */
	public Set<Individual> getEvKeys(){
		return (Set<Individual>) getEvents().keySet();
	}
	
	/**
	 * @return a TreeSet<Individual> as the this Concept ULineInd set.
	 */
	public Set<Individual> getUlKeys(){
		return (Set<Individual>) getULineMap().keySet();
	}

	/* (non-Javadoc)
	 * @see reasoner.FuzzySet#getLambda()
	 */
	/**
	 * Gives Lambda, the alpha values set, for the EventInds only
	 * @return Lambda as a TreeSet<Double>
	 */
	public TreeSet<Double> getLambda() {
		TreeSet<Double> l = new TreeSet<Double> (this.getEvents().values());		
		l.add(1.0);
		return l;
	}
	
	/**
	 * Gets the union set of alpha-values belonging to this Concept and to a 
	 * given other Concept, for the EventInds only
	 * @param c the Concept
	 * @return the union Lambda as a TreeSet<Double>
	 */
	public TreeSet<Double> totalLambda(Concept c) {
		TreeSet<Double> l = getLambda();
		if (c!=null){
			l.addAll(c.getLambda());
		}		
		return l;
	}
	
	/**
	 * Gives Lambda, the alpha values set, for the ULineInds only
	 * @return Lambda as a TreeSet<Double>
	 */
	public TreeSet<Double> getUlLambda() {
		TreeSet<Double> l = new TreeSet<Double> (this.uLineMap.values());		
		l.add(1.0);
		return l;
	}
	
	/**
	 * Gets the total set of alpha-values belonging to this Concept and to
	 * a given other concept, for the ULine Individuals only
	 * @param c the given Concept
	 * @return Lambda as a TreeSet<Double>
	 */
	public TreeSet<Double> totalUlLambda(Concept c) {
		TreeSet<Double> l = getUlLambda();
		if (c!=null){
			l.addAll(c.getUlLambda());
		}		
		return l;
	}
	
	/**
	 * The intersection between this Concept and a given second one
	 * @param c the given second Concept
	 * @return a new Concept intersection 
	 */
	public Concept intersection(Concept c) {
		Concept inter = new Concept(this.getName()+"Inter"+c.getName(),getKb());
		inter.print();
		TreeSet<Individual> keys= new TreeSet<Individual>(getEvKeys());
		Individual key;
		Double aVal, cVal, minVal;		
		Iterator<Individual> it=keys.iterator();
		while (it.hasNext()){
			key=it.next();		
			System.out.println(((Individual)key).getName());
			if (c.containsKey(key)&& this.containsKey(key)){		
				aVal=(Double)this.get(key);	
				cVal=(Double)c.get(key);
				minVal=new Double(Math.min(aVal,cVal));	
//				if (minVal!=0){					//anche se val = 0->poniamo la cardinalit=0
				inter.add((Individual)key, minVal);	
				inter.getEvents().put((Individual)key, minVal);	
//				}
			}			
		}
		inter.print();
		inter.printUlines();
		return inter;
	}
	
	/**
	 * The union between this Concept and a given second one
	 * @param c the given second Concept
	 * @return a new Concept union 
	 */
	public Concept union(Concept c) {
		Concept union = new Concept(this.getName()+"Union"+c.getName(),getKb());
		union.print();
		TreeSet<Individual> keys= new TreeSet<Individual>(getEvKeys());
		Individual key;
		Double aVal, cVal, minVal;		
		Iterator<Individual> it=keys.iterator();
		while (it.hasNext()){
			key=it.next();		
			System.out.println(((Individual)key).getName());
			if (c.containsKey(key)&& this.containsKey(key)){		
				aVal=(Double)this.get(key);	
				cVal=(Double)c.get(key);
				minVal=new Double(Math.max(aVal,cVal));	
//				if (minVal!=0){					//anche se val = 0->poniamo la cardinalit=0
				union.add((Individual)key, minVal);	
				union.getEvents().put((Individual)key, minVal);	
//				}
			}			
		}

		union.print();
		union.printUlines();
		return union;
	}
	
	/**
	 * The until between this Concept and a given second one
	 * @param c the given second Concept
	 * @return a new Concept until 
	 */
	public Concept until(Concept c) {
		Concept until = new Concept(this.getName()+"Until"+c.getName(),getKb());
		HashMap<String, ULineEvent> uLines= getKb().getULineEvents();
		Set<String> ulNames = uLines.keySet();	
		Individual xLast;
		ULineInd xo;
		ULineEvent ul;
		Double val;		
		for (String name : ulNames){
			ul = (ULineEvent) uLines.get(name);
			xLast = ul.getLast();
			xo= ul.getMyULine();	
			val = this.until(c, xLast);
			until.add(xo, val);
		}
		until.print();
		until.printUlines();
		return until;
	}
	
	/**
	 * The since between this Concept and a given second one
	 * @param c the given second Concept
	 * @return a new Concept since 
	 */
	public Concept since(Concept c) {
		Concept since = new Concept(this.getName()+"Since"+c.getName(),getKb());
		HashMap<String, ULineEvent> uLines= getKb().getULineEvents();
		Set<String> ulNames = uLines.keySet();	
		Individual xLast;
		ULineInd xo;
		ULineEvent ul;
		Double val;		
		for (String name : ulNames){
			ul = (ULineEvent) uLines.get(name);
			xLast = ul.getLast();
			xo= ul.getMyULine();	
			val = this.since(c, xLast);
			since.add(xo, val);
		}
		since.print();
		since.printUlines();
		return since;
	}
	
/**
 * Creates the new Concept corresponding to the logical expression (QT)C, where QT is a temporal 
 * quantifier and C is the present Concept
 * @param q the temporal quantifier
 * @return the (QT)C Concept
 */
public Concept tempQ(Quantifier q){
	Concept tqC = new Concept("Temp_" + q.getName() + this.getName(), getKb());
	HashMap<String, ULineEvent> uLines= getKb().getULineEvents();
	Set<String> ulNames = uLines.keySet();	
	Role event = new Role ("event", getKb());
	ULineInd xo;
	ULineEvent ul;
	Double val;		
	Set<Individual> evInds = this.getEvKeys();
	for (Individual evInd : evInds){
		event.add(evInd, 1.0);
	}
	for (String name : ulNames){
		ul = (ULineEvent) uLines.get(name);
		xo= ul.getMyULine();			
		val = q.eval(event, this.intersection(ul));
		tqC.add(xo, val);		
	}
	tqC.print();
	tqC.printUlines();
	return tqC;
}

	/**
	 * Creates the new Concept corresponding to the logical expression SubsNext(C), where
	 * C is the present Concept
	 * @param fuzzy a boolean, set to true if fuzzy subsumption is needed
	 * @return the SubsNext(C) Concept
	 */
	public Concept subsNext(boolean fuzzy){
		Concept next, inter, subsNext;		
		HashMap<String, ULineEvent> uLines = getKb().getULineEvents();
		HashSet<String> ulNames = new HashSet<String>();
		ULineEvent ul, ul1;
		Individual xlast;
		ULineInd xo;
		Double newVal = 0.0;
		subsNext = new Concept("SubsNext"+ this.getName(), getKb());
		ulNames.addAll((Set<String>)uLines.keySet());//		
		for (String ulName: ulNames){
			ul = (ULineEvent) uLines.get(ulName);
			xlast = ul.getLast();	
			xo = ul.getMyULine();
			next= this.getNextConcept(xlast);
			ul1 = ul.getWithoutLast();
			inter = this.intersection(ul1);
			newVal = inter.subsumed(next, fuzzy);
			getKb().getULineEvents().remove(ul1.getName());
			subsNext.add(xo, newVal);
		}
		return subsNext;
	}
	
	/**
	 * Creates the new Concept corresponding to the logical expression SubsPrev(C), where
	 * C is the present Concept
	 * @param fuzzy a boolean, set to true if fuzzy subsumption is needed
	 * @return the SubsPrev(C) Concept
	 */
	public Concept subsPrev(boolean fuzzy){
		Concept next, inter, subsPrev;
		HashMap<String, ULineEvent> uLines = getKb().getULineEvents();
		HashSet<String> ulNames = new HashSet<String>();
		ULineEvent ul, ul1;
		Individual xlast;
		ULineInd xo;
		Double inverseVal = 0.0;
		subsPrev = new Concept("SubsPrev"+ this.getName(), getKb());
		ulNames.addAll((Set<String>)uLines.keySet());//		
		for (String ulName: ulNames){
			ul = (ULineEvent) uLines.get(ulName);
			xlast = ul.getLast();	
			xo = ul.getMyULine();
			next= this.getNextConcept(xlast);
			ul1 = ul.getWithoutLast();
			inter = this.intersection(ul1);
			inverseVal = next.subsumed(inter, fuzzy);
			getKb().getULineEvents().remove(ul1.getName());
			subsPrev.add(xo, inverseVal);
		}
		return subsPrev;
	}
	
	/**
	 * Creates two new Concepts corresponding to the logical expressions SubsNext(C) and
	 * SubsPrev(C), where C is the present Concept
	 * @param fuzzy a boolean, set to true if fuzzy subsumption is needed
	 * @return an ArrayList<Concept>, containing the SubsNext(C) Concept first, at index 0, and
	 * the SubsPrev(C) Concept last, at 1 index
	 */
	public ArrayList<Concept> subsNextPrev(boolean fuzzy){
		Concept next, inter, subsNext, subsPrev;
		ArrayList<Concept> snp = new ArrayList<Concept>();
		HashMap<String, ULineEvent> uLines = getKb().getULineEvents();
		HashSet<String> ulNames = new HashSet<String>();
		ULineEvent ul, ul1;
		Individual xlast;
		ULineInd xo;
		Double newVal = 0.0, inverseVal = 0.0;
		subsNext = new Concept("SubsNext("+ this.getName() + ")", getKb());
		subsPrev = new Concept("SubsPrev("+ this.getName() + ")", getKb());
		ulNames.addAll((Set<String>)uLines.keySet());//		
		for (String ulName: ulNames){
			ul = (ULineEvent) uLines.get(ulName);
			xlast = ul.getLast();	
			xo = ul.getMyULine();
			next= this.getNextConcept(xlast);
			ul1 = ul.getWithoutLast();
			inter = this.intersection(ul1);
			newVal = inter.subsumed(next, fuzzy);
			inverseVal = next.subsumed(inter, fuzzy);
			getKb().getULineEvents().remove(ul1.getName());
			subsNext.add(xo, newVal);
			subsPrev.add(xo, inverseVal);
		}
		snp.add(subsNext);
		snp.add(subsPrev);
		return snp;
	}
	/**
	 * Creates two new Concepts corresponding to the logical expressions SubsNext(C) and
	 * SubsPrev(C), where C is the present Concept, and consequently creates the three Tendency
	 * Concepts Incr(C), Decr(C) and Const(C).
	 * @param fuzzy a boolean, set to true if fuzzy subsumption is needed	
	 */
	public void createTendencyConcepts(boolean fuzzy) {
		ArrayList<Concept> ar = this.subsNextPrev(fuzzy);
		HashMap<String, ULineEvent> uls = getKb().getULineEvents();
		Set<String> ulNames = uls.keySet();
		Concept next, prev;
		next = ar.get(0);
		prev = ar.get(1);
		Concept incr, decr, constant;
		incr = new Concept("Incr(" + this.getName() + ")", getKb());
		decr = new Concept("Decr(" + this.getName() + ")", getKb());
		constant = new Concept("Const(" + this.getName()+ ")", getKb());
		ULineEvent ul;
		ULineInd xo;
		double nextval, prevval;		
		for (String name : ulNames){
			ul = uls.get(name);
			xo = ul.getMyULine();
			nextval = next.get(xo);
			prevval = prev.get(xo);
			if (nextval > prevval)
				incr.add(xo, nextval - prevval);
			else if (nextval < prevval)
				decr.add(xo, prevval - nextval);
			else if (nextval == prevval)
				constant.add(xo, nextval);			
		}
	}
	
	/**
	 * Generates a linear interpolation for individuals and values in this Concept
	 * @return a String representation for the results
	 */
	public String linearInterpolation2String() {
		Point linInterpol;
		HashMap<String, ULineEvent> uLines = getKb().getULineEvents();
		Set<String> ulNames = new HashSet<String>();
		 ulNames.addAll((Set<String>)uLines.keySet());//
		ULineEvent ul;
		ULineInd xo;
		Concept inter;
		String s = "  linear interpolation: {";
		for (String name : ulNames){
			ul = (ULineEvent) uLines.get(name);
			xo = ul.getMyULine();
			inter = this.intersection(ul);
			linInterpol = inter.interpole();
			s += name + ": angCoeff = " + formatter.format(linInterpol.getX()) + "; ";
		}
		if (s.contains(";"))
			s = s.substring(0, s.lastIndexOf(";"));
		s += "}\n";	
	return s;
	}
	
	/**
	 * Tests on existence of a given Individual, as EventInd or ULineInd or StaticInd, in this Concept
	 * @param key the Individual
	 * @return true if key is in, otherwise false
	 */
	public boolean containsKey(Individual key) {
		if (key instanceof EventInd)
			return getEvents().containsKey(key);
		if (key instanceof ULineInd)
			return getULineMap().containsKey(key);
		if (key instanceof StaticInd)
			return getStaticMap().containsKey(key);
		return false;
	}
	
	/**
	 * Tests on existence of a given EventInd name in this Concept
	 * @param name the name as String
	 * @return true if any EventInd with this name is in, false otherwise
	 */
	public boolean containsIndivName(String name) {
		Iterator it = getEvKeys().iterator();
		while (it.hasNext()){
			if (((Individual)it.next()).getName().equals(name))
				return true;
		}
		return false;
	}

	/**
	 * Gives the membership value of the given Individual, 0.0 if it does not exist
	 * @param key the Individual
	 * @return the membership value
	 */
	public double get(Individual key) {
		if (getEvents().containsKey(key))
			return getEvents().get(key);
		else if (getULineMap().containsKey(key))
			return getULineMap().get(key);
		else if (getStaticMap().containsKey(key))
			return getStaticMap().get(key);
		else return 0.0;
	}
	
	/**
	 * Gives the subsumption value for this Concept in a given second one
	 * @param c the subsuming Concept 
	 * @return the value as double
	 */
	public double subsumed(Concept c, boolean fuzzy){
		if (fuzzy)
			return subsumed(new Quantifier("AlmostAll", getKb()),c); //fuzzy subsumption
		return subsumed(new Quantifier("All", getKb()),c);	//crisp subsumption


	}

	/**
	 * Returns the quantified subsumption value for this Concept in a given second one,
	 * given the quantifier.
	 * @param q the Quantifier
	 * @param c the subsuming Concept 
	 * @return the value as double
	 */
	public double subsumed(Quantifier q, Concept c){
		double sigma=0.0;
		if (q.getMaxval() <= 1.0){
			TreeSet<Double> lambda; 
			Concept ac = intersection(c);
			System.out.println("ac: ");
			ac.print();
			ac.printUlines();
			lambda = totalLambda(ac);
			System.out.println("lambda: ");
			q.print(lambda);
			Vector<Double> v = new Vector<Double>(lambda);	
			ListIterator<Double> iter=  v.listIterator(v.indexOf(v.lastElement())+1);
			Double alfa1, alfa2, ratio, qvalue, diff;	
			Integer cardAcAlfa, cardAalfa;
			while(iter.hasPrevious()){
				alfa1 =(Double) iter.previous();
				cardAcAlfa= ac.card(alfa1 ,"any");
				System.out.println("card di "+ac.getName()+" (con alfa= "+ alfa1+"): "+cardAcAlfa);				 
				cardAalfa= card(alfa1 , "any");
				System.out.println("card di "+getName()+" (con alfa= "+ alfa1+"): "+cardAalfa);
				if (cardAcAlfa==0 && cardAalfa==0)											//
					ratio = 1.0;															//poniamo 0/0=1
				else
					ratio = cardAcAlfa.doubleValue()/cardAalfa.doubleValue();
				System.out.printf("rapporto (con alfa= %.2f): %.2f \n",alfa1,ratio);
				qvalue=q.qvalue(ratio);
				System.out.printf("qvalue : %.2f \n", qvalue);
				if (iter.hasPrevious()){
					alfa2=(Double) iter.previous();
					iter.next();//per tornare all'alfa corrente
				}else{
					alfa2=0.0;
				}
				System.out.printf("alfa1 : %.2f \n", alfa1);
				System.out.printf("alfa2 : %.2f \n", alfa2);
				diff=alfa1-alfa2;
				System.out.printf("diff : %.2f \n", diff);
				sigma += diff*qvalue;
				System.out.printf("sigma : %.2f \n", sigma);				
			}
		}
		return sigma;
	}

	/** 
	 * @return a String representation for this Concept, with all EventInds and values
	 */
	public String toString(){
		String s = "";
		Individual key;
		Double value;	
		Set<Individual> keys = getEvKeys();
		Iterator<Individual> iter= keys.iterator();
		s += "Concept: " + getName() + ": {";
		while(iter.hasNext()){
			key =(Individual)iter.next();
			value= (Double)get(key);	
			s += "<"+key.getName() + ", " + formatter.format(value) + ">, ";	
		}
		if (s.contains(","))
			s = s.substring(0, s.lastIndexOf(","));
		s += "}\n";		
		return s;
	}	
	
	/** 
	 * Prints on the standard output a String representation for this Concept, 
	 * with all EventInds and values.
	 */
	public  void print() {
		System.out.println(this.toString());
	}
	/** 
	 * Prints on the standard output a String representation for this Concept, 
	 * with all ULineInds and values.
	 */
	public  void printUlines() {
		System.out.println(this.toStringUl());
	}	

	
	/**
	 * A String representation for this Concept, 
	 * with all ULineInds and values.
	 * @return the String representation
	 */
	public  String toStringUl() {
		String s = "";
		Set<Individual> keys = getULineMap().keySet();
		Iterator<Individual> iter= keys.iterator();
		Individual key;
		s += "Concept: " + getName()+ ". ULineInds: {";		
		while(iter.hasNext()){
			key =(ULineInd)iter.next();
			Double value= (Double)get(key);	
			s += "<"+key.getName() + ", " + formatter.format(value) + ">, ";	
		}
		if (s.contains(","))
			s = s.substring(0, s.lastIndexOf(","));
		s += "}\n";		
		return s;
	}	
	
	/**
	 * A String representation for this Concept, 
	 * with all ULineInds and values.
	 * @return the String representation
	 */
	public  String toNormalStringUl() {
		String s = "";
		Set<Individual> keys = getULineMap().keySet();
		Iterator<Individual> iter= keys.iterator();
		Individual key;
		s += "  " + getName()+ ": {";		
		while(iter.hasNext()){
			key =(ULineInd)iter.next();
			Double value= (Double)get(key);	
			String name = key.getName();
			name = name.substring(0,name.length()-1);
			s += "<"+ name + ", " + formatter.format(value) + ">, ";	
		}
		if (s.contains(","))
			s = s.substring(0, s.lastIndexOf(","));
		s += "}\n";		
		return s;
	}	
	/**
	 * Returns a new Concept, corresponding to the logical Concept Next(C), where C is this Concept,
	 * for Individuals preceeding the given Individual. If the given Individual is ULineInd instance, 
	 * then Next(C) is calculated for all the Individuals in C.
	 * @param ind the Individual
	 * @return the Next(C) Concept
	 */
	public Concept getNextConcept(Individual ind){
		ULineEvent ul = ((DynamicInd)ind).getULineEvent();
		Concept next;
		double time,to;
		if (ind instanceof ULineInd)
			time = Double.MAX_VALUE;
		else
			time = ind.getTime();
		Individual xo, x1;
		xo = ul.getFirst();
		to = xo.getTime();
		x1 = ul.getNext(xo);		
		next = new Concept ("next"+this.getName(), getKb());
		while (null!=x1 && to <= time){
			next.add(xo, this.get(x1));
			xo = x1;
			to = xo.getTime();
			x1 = ul.getNext(xo);			
		}
//		next.add(xo, this.get(xo));		//prima possibilit: l'ultimo elemento  introdotto col valore inalterato
//		this.getMap().remove(xo);	//altra possibilit: eliminare l'ultimo elemento
		return next;
	}

	/** 
	 * @return the events
	 */
	public Map<Individual, Double> getEvents() {
		return events;
	}

	/**
	 * @param ev the events Map to set
	 */
	public void setEvents(Map<Individual, Double> ev) {
		this.events = ev;
	}	

	/**
	 * @return the uLineMap
	 */
	public Map<Individual, Double> getULineMap() {
		return uLineMap;
	}

	/**
	 * @param lineMap the uLineMap to set
	 */
	public void setULineMap(Map<Individual, Double> lineMap) {
		uLineMap = lineMap;
	}


	/**
	 * @return the kb
	 */
	public static KnowledgeBase getKb() {
		return kb;
	}

	/**
	 * @param kkb the kb to set
	 */
	public static void setKb(KnowledgeBase kkb) {
		kb = kkb;
		setNOW(kb.getNOW());
		setSELF(kb.getSELF());
	}

	/**
	 * @return the NOW
	 */
	public static double getNOW() {
		return NOW;
	}

	/**
	 * @param now the NOW to set
	 */
	public static void setNOW(double now) {
		NOW = now;
	}

	/**
	 * @return the SELF
	 */
	public static double getSELF() {
		return SELF;
	}

	/**
	 * @param self the SELF to set
	 */
	public static void setSELF(double self) {
		SELF = self;
	}

	/**
	 * @return the sometimeMap
	 */
	public Map<Individual, Double> getSometimeMap() {
		return sometimeMap;
	}

	/**
	 * @param sometimeMap the sometimeMap to set
	 */
	public void setSometimeMap(Map<Individual, Double> sometimeMap) {
		this.sometimeMap = sometimeMap;
	}

	/**
	 * @return the alwaysMap
	 */
	public Map<Individual, Double> getAlwaysMap() {
		return alwaysMap;
	}

	/**
	 * @param alwaysMap the alwaysMap to set
	 */
	public void setAlwaysMap(Map<Individual, Double> alwaysMap) {
		this.alwaysMap = alwaysMap;
	}

	/**
	 * @return the staticMap
	 */
	public Map<Individual, Double> getStaticMap() {
		return staticMap;
	}

	/**
	 * @param staticMap the staticMap to set
	 */
	public void setStaticMap(Map<Individual, Double> staticMap) {
		this.staticMap = staticMap;
	}
}
