/**
 * Temporal Quantified fuzzy DL's
 */
package reasoner;

import java.text.NumberFormat;
import java.util.ListIterator;
import java.util.TreeSet;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This class represents a FuzzySet having trapezoidal membership function, delimited by 4 point: 
 * first and last having 0.0 value, second and third having  1.0 value. 
 * @author marco furlan
 */
public class Trapezoid implements FuzzySet{
	//the 4 vertices: a and d have value 0.0; b and c have value 1.0
	private Double a;
	//the 4 vertices: a and d have value 0.0; b and c have value 1.0
	private Double b;
	//the 4 vertices: a and d have value 0.0; b and c have value 1.0
	private Double c;
	//the 4 vertices: a and d have value 0.0; b and c have value 1.0
	private Double d;
	protected double maxval = Double.MAX_VALUE; 
	protected double minval = Double.NEGATIVE_INFINITY; 
	protected String name = "Trapezoid";
	protected String meaning = " ";
	protected static final String patternAround="Around\\((-?[0-9]+.[0-9]+)\\)";
	protected static final String patternBetween="Between\\((-?[0-9]+.[0-9]+)\\)" +
			"and\\((-?[0-9]+.[0-9]+)\\)";
	private static double v1;
	private static double v2;
	protected NumberFormat formatter; 	
	
	/**
	 * constructor for the null trapezoid
	 */
	public Trapezoid (){				
		setVertices(0.0,0.0,0.0,0.0);
		formatter = NumberFormat.getNumberInstance();
		formatter.setMaximumFractionDigits(2);
	}
	
	/**
	 * Constructor by the 4 vertices
	 * @param a1 the first vertice
	 * @param b1 the second vertice
	 * @param c1 the third vertice
	 * @param d1 the fourth vertice
	 */
	public Trapezoid (Double a1,Double b1,Double c1,Double d1){				
		this();
		setVertices(a1,b1,c1,d1);		
	}
	/**
	 * Set the vertices
	 * @param a1 the first vertice
	 * @param b1 the second vertice
	 * @param c1 the third vertice
	 * @param d1 the fourth vertice
	 */
	protected void setVertices(Double a1,Double b1,Double c1,Double d1){
		seta(Math.min(a1, d1));
		setb(Math.min(b1, c1));
		setc(Math.max(b1, c1));
		setd(Math.max(a1, d1));		
		fixVertices();
	}
	/**
	 * If the given name matches with "Around(...)", extracts the referred value and
	 * return true
	 * @param name the name
	 * @return true if name matches, false otherwise
	 */
	public static boolean getAroundValue(String name){
		Pattern p = Pattern.compile(patternAround, Pattern.CASE_INSENSITIVE);
		Matcher m=p.matcher(name);
		if (m.find()){
			v1=Double.parseDouble(m.group(1).replace(",", "."));		
			return true;
		}else return false;
	}
	/**
	 * If the given name matches with "Between(...)and(...)", extracts the two referred values
	 * and return true
	 * @param name the name
	 * @return true if name matches, false otherwise
	 */
	public static boolean getBetweenValue(String name){
		Pattern p = Pattern.compile(patternBetween, Pattern.CASE_INSENSITIVE);
		Matcher m=p.matcher(name);
		if (m.find()){
			v1=Double.parseDouble(m.group(1).replace(",", "."));
			v2=Double.parseDouble(m.group(2).replace(",", "."));
			return true;
		}else
		return false;
	}
	
	/**
	 * Constructor by name: configures the vertices, knowing this Trapezoid name
	 * @param name the name
	 */
	public Trapezoid(String name) {
		this();	
		setName(name);
		//some important cases
		if (name.trim().equalsIgnoreCase("One")){				
			setVertices(minval,minval,maxval,maxval);
			setName("One");
			setMeaning("One");
		}else
			if (name.trim().equalsIgnoreCase("All")){		
				setVertices(1.0,1.0,1.0,1.0);
				setName("All");
				setMeaning("All");
			}else
				if (name.trim().equalsIgnoreCase("None")){		
					setVertices(0.0,0.0,0.0,0.0);
					setName("None");
					setMeaning("None");
				}else					
		if (getAroundValue(name.trim())){
			if(v1>1.0||v1<-1.0)
			setVertices(v1-1.0,v1,v1,v1+1.0);	
			else
				setVertices(v1-0.1,v1,v1,v1+0.1);
//			setName("Around("+ v1 + ")");
			setMeaning();
			setName(getMeaning());
		}else
			if(getBetweenValue(name.trim())){
				if(v1>1.0||v1<-1.0)
				setVertices(v1-1.0,v1,v2,v2+1.0);
				else
					setVertices(v1-0.1,v1,v2,v2+0.1);
				//setName("Between("+ v1 + ")and(" + v2 + ")");
				setMeaning();
				setName(getMeaning());
			}
			
	}
	/**
	 * Sets the meaning, for some particular cases.
	 */
	protected void setMeaning() {
		//triangular shape
		if (getb().compareTo(getc())==0 && geta().compareTo(getd())!=0){
			setMeaning ("Around(" + formatter.format(getb()) + ")");
		}
		//trapezium shape
		if (getb().compareTo(getc())!=0 && geta().compareTo(getd())!=0){			
			setMeaning ("Between(" + formatter.format(getb()) + ")and(" + formatter.format(getc()) +")");			
		}
		//left shoulder shape
		if (geta()==0.0 && getb()==0.0){
			setMeaning ("LessThan("+ formatter.format(getc())+")");					
		}
	}
	
	/**
	 * @return the d vertice
	 */
	protected Double getd() {		
		return d;
	}
	/**
	 * @return the c vertice
	 */
	protected Double getc() {		
		return c;
	}
	/**
	 * @return the b vertice
	 */
	protected Double getb() {
		return b;
	}
	/**
	 * @return the a vertice
	 */
	protected Double geta() {
		return a;
	}
	/**
	 * @param d1 the new d value
	 */
	protected void setd(Double d1) {
		d=d1;		
	}
	/**
	 * @param c1 the new c value
	 */
	protected void setc(Double c1) {
		c=c1;		
	}
	/**
	 * @param b1 the new b value
	 */
	protected void setb(Double b1) {
		b=b1;		
	}
	/**
	 * @param a1 the new a value
	 */
	protected void seta(Double a1) {
		a=a1;		
	}
	
	/** 
	 * return a String representation for this Trapezoid as a sequence of vertices
	 * @see java.lang.Object#toString()
	 */
	public String toString(){
		String s = "";
		s += "Trapezoid: " + this.getName() + ": ";
		s += "   " + this.getMeaning() + ": ";
		s += "   0.0/"+formatter.format(geta())+ " + 1.0/"+formatter.format(getb())+
		" + 1.0/"+formatter.format(getc())+ " + 0.0/"+formatter.format(getd());
		s += "\n";
		return s;
	}
	/**
	 * Fuzzy Set sum between this Trapezoid and a second one 
	 * @param t the second Trapezoid
	 * @return the sum Trapezoid
	 */
	public Trapezoid sum(Trapezoid t){
		setVertices(geta()+t.geta(),getb()+t.getb(),getc()+t.getc(), getd()+t.getd());
		return this;
	}
	/**
	 * Fuzzy Set difference between this Trapezoid and a second one 
	 * @param t the second Trapezoid
	 * @return the difference Trapezoid
	 */
	public Trapezoid diff(Trapezoid t){
		setVertices(geta()-t.getd(),getb()-t.getc(),getc()-t.getb(), getd()-t.geta());
		return this;
	}
	
	/**
	 * Fuzzy Set product between this Trapezoid and a second one 
	 * @param t the second Trapezoid
	 * @return the product Trapezoid
	 */
	public Trapezoid prod(Trapezoid t){
		setVertices(Math.min(Math.min(geta()*t.geta(),geta()*t.getd()),
				Math.min(getd()*t.geta(), getd()*t.getd())),
				Math.min(Math.min(getb()*t.getb(),getb()*t.getc()),
						Math.min(getc()*t.getb(), getc()*t.getc())),
				Math.max(Math.max(getb()*t.getb(),getb()*t.getc()),
						Math.max(getc()*t.getb(), getc()*t.getc())),
				Math.max(Math.max(geta()*t.geta(),geta()*t.getd()),
						Math.max(getd()*t.geta(), getd()*t.getd())));
		return this;
	}
	
	/** Fuzzy Set ratio between this Trapezoid and a second one. 
	 *  Possible for divisor Trapezoid not zero containing only!
	 * @param t the second Trapezoid
	 * @return the ratio Trapezoid
	 */
	public Trapezoid ratio(Trapezoid t){
		if(t.geta()!=0 && t.getb()!=0 && t.getc()!=0 && t.getd()!=0){
		setVertices(Math.min(Math.min(geta()/t.geta(),geta()/t.getd()),
				Math.min(getd()/t.geta(), getd()/t.getd())),
				Math.min(Math.min(getb()/t.getb(),getb()/t.getc()),
						Math.min(getc()/t.getb(), getc()/t.getc())),
				Math.max(Math.max(getb()/t.getb(),getb()/t.getc()),
						Math.max(getc()/t.getb(), getc()/t.getc())),
				Math.max(Math.max(geta()/t.geta(),geta()/t.getd()),
						Math.max(getd()/t.geta(), getd()/t.getd())));
		}else{
			System.err.println("divisor Trapezoid contains 0.0: impossible ratio!");
		}
		return this;
	}
	/**
	 * Fixes vertices to be internal to boundary min and max values
	 */
	protected void fixVertices(){		
		if (geta().doubleValue()< getMinval()) seta(getMinval());
		if (getb().doubleValue()< getMinval()) setb(getMinval());
		if (getc().doubleValue()< getMinval()) setc(getMinval());
		if (getd().doubleValue()< getMinval()) setd(getMinval());
		if (geta().doubleValue()> getMaxval()) seta(getMaxval());
		if (getb().doubleValue()> getMaxval()) setb(getMaxval());
		if (getc().doubleValue()> getMaxval()) setc(getMaxval());
		if (getd().doubleValue()> getMaxval()) setd(getMaxval());
	}
	
//	/**
//	 * Intersection with a FuzzySet
//	 * @param f the FuzzySet
//	 * @return the intersection 
//	 */
//	public Trapezoid intersection(FuzzySet f){
//		//to do
//		return this;
//	};
//	/**
//	 * Union with a FuzzySet
//	 * @param f the FuzzySet
//	 * @return the union 
//	 */
//	public Trapezoid union(FuzzySet f){
//		//to do
//		return this;
//	};
	/**
	 * @return a Vector<Double> containing vertices positions.
	 */
	protected Vector<Double> getKeys(){
		Vector<Double> v = new Vector<Double>();
		v.add(this.geta());
		v.add(this.getb());
		v.add(this.getc());
		v.add(this.getd());
		return v;
	}
	/**
	 * @return a Vector<Double> containing vertices values.
	 */
	protected Vector<Double> getVertValues(){
		Vector<Double> v = new Vector<Double>();
		v.add(0.0);
		v.add(1.0);
		v.add(1.0);
		v.add(0.0);
		return v;
		
	}
	
	/**
	 * Evaluation on this Trapezoid, for a given position
	 * @param d the position
	 * @return the Double value
	 */
	public Double qvalue(Double d) {
		Double key1, key2, v1, v2, qval;		
		Vector<Double> k = this.getKeys();	
		Vector<Double> val = this.getVertValues();
		ListIterator<Double> iter=  k.listIterator();
		ListIterator<Double> iter1=  val.listIterator();		
		qval=0.0;
//		System.out.printf("0_qval: %.2f \n", qval);
		while(iter.hasNext()){
			key1 =(Double) iter.next();
//			System.out.printf("key1: %.2f \n", key1);
			v1=(Double)iter1.next();
//			System.out.printf("v1: %.2f \n", v1);
			if (iter.hasNext()){
				key2 =(Double) iter.next();
//				System.out.printf("key2: %.2f \n", key2);
				v2= (Double)iter1.next();
//				System.out.printf("v2: %.2f \n", v2);
				iter.previous();
				iter1.previous();
			}else {
				key2=getMaxval();
//				System.out.printf("key2: %.2f \n", key2);
				v2=v1;
//				System.out.printf("v2: %.2f \n", v2);
			}
			if (d<key1) return qval;
			if (d.equals(key1)){
				if (key1.equals(key2)){
					qval=Math.max(v1,v2);
//					System.out.printf("1_qval: %.2f \n", qval);
					return qval;
				}else{
					qval=v1;
//					System.out.printf("2_qval: %.2f \n", qval);
					return qval;
				}
			}else if(d>key1 && d<key2)
				if(v2.equals(v1)){
					qval=v1;
//					System.out.printf("2_qval: %.2f \n", qval);
					return qval;
				}else if(v2.compareTo(v1)>0){
						qval=(d-key1)*(v2-v1)/(key2-key1) + v1;
//						System.out.printf("3_qval: %.2f \n", qval);
						return qval;
				}else{
						qval=(key2-d)*(v1-v2)/(key2-key1) + v2;
//						System.out.printf("4_qval: %.2f \n", qval);
						return qval;
				}					
			}		
		return qval;
	}
	/**
	 * @return the maxval
	 */
	public double getMaxval() {
		return maxval;
	}
	/**
	 * @param maxval the maxval to set
	 */
	public void setMaxval(double maxval) {
		this.maxval = maxval;
	}
	/**
	 * @return the minval
	 */
	public double getMinval() {
		return minval;
	}
	/**
	 * @param minval the minval to set
	 */
	public void setMinval(double minval) {
		this.minval = minval;
	}
	/**
	 * @return the name
	 */
	public String getName() {
		if (name.equals("Trapezoid"))setName(getMeaning());
		return name;
	}
	/**
	 * @param name1 the name to set
	 */
	public void setName(String name1) {
		name = name1;
	}
	/* (non-Javadoc)
	 * @see reasoner.FuzzySet#card(java.lang.Double, java.lang.String)
	 */
	/**
	 * Gives the integer cardinality at the given alpha level.
	 * The String mode operates selection over Individuals.
	 * @param alpha1 the given alpha level
	 * @param mode the selecting String
	 * @return the cardinality
	 */
	public Integer card(Double alpha1, String mode)
	{
		//to do
		return null;
	}
	/* (non-Javadoc)
	 * @see reasoner.FuzzySet#getLambda()
	 */
	/**
	 * Gives Lambda, the alpha level set for membership values
	 * @return Lambda as a TreeSet<Double>
	 */
	public TreeSet<Double> getLambda()
	{
		// to do
		return null;
	}
	/**
	 * Prints on the standard output a String representation for this
	 * Trapezoid
	 */
	protected void print() {		
		System.out.println("name: "+ getName());
		System.out.println("meaning: "+ getName());
		System.out.println("vertices:");
			System.out.printf("%.2f --> 0.0;  ",geta());
			System.out.printf("%.2f --> 1.0;  ",getb());
			System.out.printf("%.2f --> 1.0;  ",getc());
			System.out.printf("%.2f --> 0.0;  ",getd());
			System.out.printf("minVal: %.2f ;  ",getMinval());
			System.out.printf("maxVal: %.2f ;  ",getMaxval());
		System.out.println("");
	}
	/**
	 * @return the meaning
	 */
	public String getMeaning() {
		setMeaning();		
		return meaning;
	}
	/**
	 * @param meaning the meaning to set
	 */
	public void setMeaning(String meaning) {
		this.meaning = meaning;
	}
	
}
