/**
 * Temporal Quantified fuzzy DL's
 */
package reasoner;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;

import tbox.*;

/**
 * This class represents a knowledge base as a container of Concepts, ConcreteConcepts,
 * Individuals, ULineEvents, Roles, Quantifiers and one Query
 * @author marco furlan
 *
 */
public class KnowledgeBase implements Serializable{
	/**
	 * The Concept map
	 */
	private HashMap<String, Concept> concepts;
	/**
	 * The ConcreteConcept map
	 */
	private HashMap<String, ConcreteConcept> concrConcepts;
	/**
	 * The Individual map
	 */
	private HashMap<String, Individual> individuals;
	/**
	 * The ULineEvent map
	 */
	private HashMap<String, ULineEvent> uLineEvents;
	/**
	 * The role map
	 */
	private HashMap<String, Role> roles;
//	private HashMap assertions;	
	/**
	 * The Quantifier map
	 */
	private HashMap<String, Quantifier> quantifiers;
	/**
	 * The Query
	 */
	private Query query;
	/**
	 * The NOW constant
	 */
	private final double NOW = 0.0;
	/**
	 * The SELF constant
	 */
	private final double SELF = 9.9;
	
	/**
	 * Constructor for a new empty Knowledge Base as a container of empty HashMaps of Concepts,
	 * ConcreteConcepts, Individuals, ULineEvents, Roles and Quantifiers and with a null Query
	 */
	public KnowledgeBase() {
		this.setConcepts( new HashMap<String, Concept>());
		this.setConcrConcepts ( new HashMap<String, ConcreteConcept>());
		this.setIndividuals ( new HashMap<String, Individual>());
		this.setULineEvents (new HashMap<String, ULineEvent>());
		this.setRoles (new HashMap<String, Role>());
//		this.assertions = new HashMap();		
		this.setQuantifiers (new HashMap<String, Quantifier>());
		this.setQuery(null);
	}
	/**
	 * Constructor for a new Knowledge Base from given HashMaps of elements
	 * @param concepts the HashMap of Concept names and Concepts
	 * @param concrConcepts the HashMap of ConcreteConcept names and ConcreteConcepts
	 * @param individuals the HashMap of Individual names and Individuals
	 * @param uLineEvents the HashMap of ULineEvent names and ULineEvents
	 * @param roles	 the HashMap of Role names and Roles	 
	 * @param quantifiers the HashMap of Quantifier names and Quantifiers
	 */
	public KnowledgeBase(HashMap<String, Concept> concepts, HashMap<String, ConcreteConcept> concrConcepts,
			HashMap<String, Individual> individuals, HashMap<String, ULineEvent> uLineEvents, 
			HashMap<String, Role> roles,HashMap<String, Quantifier> quantifiers) {
		this.setConcepts(concepts);
		this.setConcrConcepts(concrConcepts);
		this.setIndividuals(individuals);
		this.setULineEvents(uLineEvents);
		this.setRoles(roles);		
//		this.setAssertions (assertions);		
		this.setQuantifiers(quantifiers);
		this.setQuery(null);
	}
	/**
	 * Constructor for a new Knowledge Base from given HashMaps of elements
	 * @param concepts the HashMap of Concept names and Concepts
	 * @param concrConcepts the HashMap of ConcreteConcept names and ConcreteConcepts
	 * @param individuals the HashMap of Individual names and Individuals
	 * @param uLineEvents the HashMap of ULineEvent names and ULineEvents
	 * @param roles	 the HashMap of Role names and Roles	 
	 * @param quantifiers the HashMap of Quantifier names and Quantifiers
	 * @param query the given Query
	 */
	public KnowledgeBase(HashMap<String, Concept> concepts, HashMap<String, ConcreteConcept> concrConcepts, 
			HashMap<String, Individual> individuals,HashMap<String, ULineEvent> uLineEvents,
			HashMap<String, Role> roles,  HashMap<String, Quantifier> quantifiers, Query query) {
		this(concepts, concrConcepts, individuals, uLineEvents, roles, quantifiers);		
		this.setQuery(query);
	}
	
	/**
	 * Gives the total time set
	 * @return the time set 
	 */	
	public Set<Double> getTimeSet(){
		Set<String> uls = this.getULineEvents().keySet();
		Set<Double> times = new TreeSet<Double>();
		ULineEvent ul;
		for (String name : uls){
			ul = (ULineEvent)this.getULineEvents().get(name);
			times.addAll(ul.getTimeSet());			
		}
		return times;
	}
	
	/**
	 * Adds a Role to this KB
	 * @param r the Role
	 */
	public void addRole(Role r){
		this.getRoles().put(r.getName(), r);
	}
	
	/**
	 * Adds a Quantifier to this KB
	 * @param q the Quantifier
	 */
	public void addQuantifier(Quantifier q){
		this.getQuantifiers().put(q.getName(), q);
	}
	/**
	 * Adds a Concept to this KB
	 * @param c the concept to add
	 */
	public void addConcept(Concept c) {
		this.concepts.put(c.getName(), c);
	}
	
	/**
	 * Adds an ULineEvent to this KB
	 * @param ul the ULineEvent to add
	 */
	public void addULineEv(ULineEvent ul) {
		this.uLineEvents.put(ul.getName(), ul);
	}
	
	/**
	 * Adds an Individual to this KB
	 * @param i the Individual to add
	 */
	public void addIndividual(Individual i) {
		this.individuals.put(i.getName(), i);
	}
	
//	/**
//	 * @param a the assertion to add
//	 */
//	public void addAssertion(Assertion a) {
//		//this.assertions.put(key, value);
//	}
	
	/**
	 * Adds a ConcreteConcept to this KB
	 * @param tc the ConcreteConcept
	 */
	public void addConcreteConcept(ConcreteConcept tc) {
		this.getConcrConcepts().put(tc.getName(), tc);		
	}
	
	/**
	 * @return the HashMap of Quantifiers
	 */
	public HashMap<String, Quantifier> getQuantifiers() {
		return quantifiers;
	}
	
	/**
	 * @param quantifiers the HashMap of Quantifiers to set
	 */
	public void setQuantifiers(HashMap<String, Quantifier> quantifiers) {
		this.quantifiers = quantifiers;
	}
//	/**
//	 * @return the assertions
//	 */
//	public HashMap getAssertions() {
//		return assertions;
//	}

//	/**
//	 * @param assertions the assertions to set
//	 */
//	public void setAssertions(HashMap assertions) {
//		this.assertions = assertions;
//	}
	
	/**
	 * @return the HashMap of Concepts
	 */
	public HashMap<String, Concept> getConcepts() {
		return concepts;
	}
	
	/**
	 * @param concepts the HashMap of Concepts to set
	 */
	public void setConcepts(HashMap<String, Concept> concepts) {
		this.concepts = concepts;
	}
	
	/**
	 * @return the HashMap of Roles
	 */
	public HashMap<String, Role> getRoles() {
		return roles;
	}
	
	/**
	 * @param roles the HashMap of Roles to set
	 */
	public void setRoles(HashMap<String, Role> roles) {
		this.roles = roles;
	}
	
	/**
	 * @return the HashMap of uLineEvents in this KB
	 */
	public HashMap<String, ULineEvent> getULineEvents() {
		return uLineEvents;
	}
	
	/**
	 * @param uLineEvents the HashMap of uLineEvents to set in this KB
	 */
	public void setULineEvents(HashMap<String, ULineEvent> uLineEvents) {
		this.uLineEvents = uLineEvents;
	}	

	/**
	 * @return the concrConcepts, the HashMap of ConcreteConcepts
	 */
	public HashMap<String, ConcreteConcept> getConcrConcepts() {
		return concrConcepts;
	}

	/**
	 * sets the HashMap of ConcreteConcepts
	 * @param concrConcepts the HashMap of concrConcepts 
	 */
	public void setConcrConcepts(HashMap<String, ConcreteConcept> concrConcepts) {
		this.concrConcepts = concrConcepts;
	}

	/**
	 * @return individuals, the HashMap of Individuals
	 */
	public HashMap<String, Individual> getIndividuals() {
		return individuals;
	}

	/**
	 * @param individuals the HashMap of Individuals to set
	 */
	public void setIndividuals(HashMap<String, Individual> individuals) {
		this.individuals = individuals;
	}

	/**
	 * @return the Query
	 */
	public Query getQuery() {
		return query;
	}
	
	/**
	 * @param query the Query to set
	 */
	public void setQuery(Query query) {
		this.query = query;
	}

	/**
	 * @return the constant NOW
	 */
	public double getNOW() {
		return NOW;
	}

	/**
	 * @return the constant SELF
	 */
	public double getSELF() {
		return SELF;
	}	
	/**
	 * Clears all the maps in this KB
	 */
	public void clear(){
		getConcepts().clear();
		getConcrConcepts().clear();
		getULineEvents().clear();
		getRoles().clear();
//		getAssertions().clear();
		getIndividuals().clear();
		getQuantifiers().clear();
		setQuery(null);
	}
	/**
	 * Gives a String representation for this KB
	 * @return the String representation
	 */
	public String toString(){
		String s = "";
		Set<String> keys;
		Iterator<String> it;
//		s += "Knowledge Base \n";
		keys = concepts.keySet();
		it = keys.iterator();
		while(it.hasNext()){
			Concept c = (Concept) concepts.get(it.next());
			s += c.toString();
		}
		keys = concrConcepts.keySet();
		it = keys.iterator();
		while(it.hasNext()){
			ConcreteConcept c = (ConcreteConcept) concrConcepts.get(it.next());
			s += c.toString();
		}
		keys = quantifiers.keySet();
		it = keys.iterator();
		while(it.hasNext()){
			Quantifier q = (Quantifier) quantifiers.get(it.next());
			s += q.toString();
		}
		keys = roles.keySet();
		it = keys.iterator();
		while(it.hasNext()){
			Role r = (Role) roles.get(it.next());
			s += r.toString();
		}
//		keys = universalLines.keySet();
//		it = keys.iterator();
//		while(it.hasNext()){
//			ULineEvent ul = (ULineEvent) uLineEvents.get(it.next());
//			s += ul.toString();
//		}
		if (null!=getQuery())
			s += getQuery().toString();
		return s;
	}
	public synchronized void xmlSave() throws JAXBException, IOException{
		ObjectFactory factory = new ObjectFactory();
		TBoxType tbox = factory.createTBoxType();
		Set<String> keys;
		Iterator<String> it;
		keys = concepts.keySet();
		it = keys.iterator();
//		while(it.hasNext()){
//			Concept c = (Concept) concepts.get(it.next());
//			//todo
//		}
		keys = concrConcepts.keySet();
		it = keys.iterator();
		while(it.hasNext()){
			ConcreteConcept c = (ConcreteConcept) concrConcepts.get(it.next());
			SimpleConcept sc = factory.createSimpleConcept();			
			TrapeziumType t =factory.createTrapeziumType();
			t.setName(c.getName());
			t.setMeaning(c.getMeaning());
			t.setMinVal(c.getMinval());
			t.setMaxVal(c.getMaxval());
			t.setXA(c.geta());
			t.setXB(c.getb());
			t.setXC(c.getc());
			t.setXD(c.getd());
			sc.setTrapezium(t);
			tbox.getSimpleConcepts().add(sc);
		}
		keys = quantifiers.keySet();
		it = keys.iterator();
		while(it.hasNext()){
			Quantifier q = (Quantifier) quantifiers.get(it.next());
			QuantifierType qt = factory.createQuantifierType();
			TrapeziumType t =factory.createTrapeziumType();
			t.setName(q.getName());
			t.setMeaning(q.getMeaning());
			t.setMinVal(q.getMinval());
			t.setMaxVal(q.getMaxval());
			t.setXA(q.geta());
			t.setXB(q.getb());
			t.setXC(q.getc());
			t.setXD(q.getd());
			qt.setTrapezium(t);
			qt.setRelative(q.isRelative());
			tbox.getQuantifiers().add(qt);
		}
		keys = roles.keySet();
		it = keys.iterator();
		while(it.hasNext()){
			Role r = (Role) roles.get(it.next());
			RoleType rt = factory.createRoleType();
			rt.setName(r.getName());
			tbox.getRoles().add(rt);
		}
		keys = this.getULineEvents().keySet();
		it = keys.iterator();
//		while(it.hasNext()){
//			ULineEvent ul = (ULineEvent) this.getULineEvents().get(it.next());
//			//todo
//		}
		QueryType qt = factory.createQueryType();
		if(query!=null){
			String queryClass = query.getClass().getName();
			System.out.println("queryClass: "+ queryClass);
			if(queryClass.equals("reasoner.RoleQuery")){
				qt.setIsRoleQuery(true);	
				if(null!=((RoleQuery)query).getTempq())
					qt.setTempQuantifier(((RoleQuery)query).getTempq().getName());
				if(null!=((RoleQuery)query).getQ())
					qt.setQuantifier(((RoleQuery)query).getQ().getName());
				qt.setRole(((RoleQuery)query).getR().getName());
				qt.setConcept(((RoleQuery)query).getTc().getName());
			}else if(queryClass.equals("reasoner.OpQuery")){
				qt.setIsRoleQuery(false);	
				if(null!=((OpQuery)query).getQ())
					qt.setQuantifier(((OpQuery)query).getQ().getName());
				qt.setConcept(((OpQuery)query).getTc1().getName());
				qt.setOperator(((OpQuery)query).getOperator());
				qt.setConcept2(((OpQuery)query).getTc2().getName());
			}
			tbox.setQuery(qt);
		}
		JAXBElement<TBoxType> tboxEl = (JAXBElement<TBoxType>)factory.createTBox(tbox);
		JAXBContext ctx = JAXBContext.newInstance("tbox");
		Marshaller m = ctx.createMarshaller();		
    	m.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
    	String s = "MTBox";
    	if (query!=null)
    		s = query.getMeaning();
        FileOutputStream out = new FileOutputStream(s +".xml");
        m.marshal(tboxEl, out);
        out.flush();
        out.close();
	}

}
